/*******************************

 Rue's anti-fault monitor.

*******************************/

#include "defines.h"
#include "serial.h"
#include "flash.h"

/* Uncomment the following to save code space */
#define REMOVE_FUSE_AND_LOCK_BIT_SUPPORT
#define REMOVE_BLOCK_SUPPORT
#define REMOVE_EEPROM_BYTE_SUPPORT
#define REMOVE_FLASH_BYTE_SUPPORT


#ifndef REMOVE_BLOCK_SUPPORT
    unsigned char BlockLoad(unsigned int size, unsigned char mem, long *address);
    void BlockRead(unsigned int size, unsigned char mem, long *address);
    
    /* BLOCKSIZE should be chosen so that the following holds: BLOCKSIZE*n = PAGESIZE,  where n=1,2,3... */
    #define BLOCKSIZE PAGESIZE
#endif



void main(void) {
    long          address;
    unsigned char val;

    initbootuart(); // Initialize UART.

    while(1) {
    
        val = recchar(); // Wait for command character.

        if(0) {
        
        } else if(val=='A') { // Set address...
         // NOTE: Flash addresses are given in words, not bytes.
            address=(recchar()<<8) | recchar(); // Read address high and low byte.
            sendchar('\r'); // Send OK back.
            
        } else if(val=='e') { // Chip erase...
            for(address = 0; address < APP_END;address += PAGESIZE)
            { // NOTE: Here we use address as a byte-address, not word-address, for convenience.
                _WAIT_FOR_SPM();
                _PAGE_ERASE( address );
            }

            sendchar('\r'); // Send OK back.
        }

#ifndef REMOVE_BLOCK_SUPPORT
        // Check block load support.
        else if(val=='b')
	    {
			sendchar('Y'); // Report block load supported.
			sendchar((BLOCKSIZE>>8) & 0xFF); // MSB first.
			sendchar(BLOCKSIZE&0xFF); // Report BLOCKSIZE (bytes).
		}


        // Start block load.
		else if(val=='B')
		{
    	    temp_int = (recchar()<<8) | recchar(); // Get block size.
	    	val = recchar(); // Get memtype.
		    sendchar( BlockLoad(temp_int,val,&address) ); // Block load.
		    
	    } else if(val=='g') { // Start block read.
    	    temp_int = (recchar()<<8) | recchar(); // Get block size.
			val = recchar(); // Get memtype
    		BlockRead(temp_int,val,&address); // Block read
		}		
#endif

#ifndef REMOVE_FLASH_BYTE_SUPPORT
        // Read program memory.
        else if(val=='R')
        {
            // Send high byte, then low byte of flash word.
            _WAIT_FOR_SPM();
            _ENABLE_RWW_SECTION();
            #pragma diag_suppress=Pe1053 // Suppress warning for conversion from long-type address to flash ptr.
#ifdef LARGE_MEMORY
            sendchar( _LOAD_PROGRAM_MEMORY( (address << 2)+2 ) );
            sendchar( _LOAD_PROGRAM_MEMORY( (address << 2)+1 ) );
#else
            sendchar( _LOAD_PROGRAM_MEMORY( (address << 1)+1 ) );
            sendchar( _LOAD_PROGRAM_MEMORY( (address << 1)+0 ) );
#endif
            #pragma diag_default=Pe1053 // Back to default.

            address++; // Auto-advance to next Flash word.
        }


        // Write program memory, low byte.
        else if(val=='c')
        { // NOTE: Always use this command before sending high byte.
            temp_int=recchar(); // Get low byte for later _FILL_TEMP_WORD.
            sendchar('\r'); // Send OK back.
        }


        // Write program memory, high byte.
        else if(val=='C')
        {
            temp_int |= (recchar()<<8); // Get and insert high byte.
            _WAIT_FOR_SPM();
#ifdef LARGE_MEMORY
            _FILL_TEMP_WORD( (address << 2), temp_int ); // Convert word-address to byte-address and fill.
#else
            _FILL_TEMP_WORD( (address << 1), temp_int ); // Convert word-address to byte-address and fill.
#endif
            address++; // Auto-advance to next Flash word.
            sendchar('\r'); // Send OK back.
        }


        // Write page.
        else if(val== 'm')
        {
            if( address >= (APP_END>>1) ) // Protect bootloader area.
            {
                sendchar('?');
            } else
            {
                _WAIT_FOR_SPM();
#ifdef LARGE_MEMORY
                _PAGE_WRITE( address << 2 ); // Convert word-address to byte-address and write.
#else
                _PAGE_WRITE( address << 1 ); // Convert word-address to byte-address and write.
#endif
            }

            sendchar('\r'); // Send OK back.
        }
#endif

#ifndef REMOVE_EEPROM_BYTE_SUPPORT
        // Write EEPROM memory.
        else if (val == 'D')
        {
            _WAIT_FOR_SPM();
            EEARL = address; // Setup EEPROM address.
            EEARH = (address >> 8);
            EEDR = recchar(); // Get byte.
            EECR |= (1<<EEMWE); // Write byte.
            EECR |= (1<<EEWE);
            while (EECR & (1<<EEWE)) // Wait for write operation to finish.
                ;

            address++; // Auto-advance to next EEPROM byte.
            sendchar('\r');// Send OK back.
        }


        // Read EEPROM memory.
        else if (val == 'd')
        {
            EEARL = address; // Setup EEPROM address.
            EEARH = (address >> 8);
            EECR |= (1<<EERE); // Read byte...
            sendchar(EEDR); // ...and send it back.
            address++; // Auto-advance to next EEPROM byte.
        }
#endif

#ifndef REMOVE_FUSE_AND_LOCK_BIT_SUPPORT
        // Write lockbits.
        else if(val=='l')
        {
            _WAIT_FOR_SPM();
            _SET_LOCK_BITS( recchar() ); // Read and set lock bits.
            sendchar('\r'); // Send OK back.
        }


        // Read lock bits.
        else if(val=='r')
        {
            _WAIT_FOR_SPM();
            sendchar( _GET_LOCK_BITS() );
        }


        // Read fuse bits.
        else if(val=='F')
        {
            _WAIT_FOR_SPM();
            sendchar( _GET_LOW_FUSES() );
        }


        // Read high fuse bits.
        else if(val=='N')
        {
            _WAIT_FOR_SPM();
            sendchar( _GET_HIGH_FUSES() );
        }


        // Read extended fuse bits.
        else if(val=='Q')
        {
            _WAIT_FOR_SPM();
            sendchar( _GET_EXTENDED_FUSES() );
        }
#endif

        // Exit bootloader.
        else if(val=='E')   {
            _WAIT_FOR_SPM();
            _ENABLE_RWW_SECTION();
            sendchar('\r');
        }

        // The last command to accept is ESC (synchronization).
        else if(val!=0x1b)                  // If not ESC, then it is unrecognized...
        {
            sendchar('?');
        }
    } // end: for(;;)
   
} // end: main


#ifndef REMOVE_BLOCK_SUPPORT
unsigned char BlockLoad(unsigned int size, unsigned char mem, long *address)
{
    unsigned char buffer[BLOCKSIZE];
    unsigned int data;
    long tempaddress;
	
    // EEPROM memory type.
    if(mem=='E')
    {
        /* Fill buffer first, as EEPROM is too slow to copy with UART speed */
        for(tempaddress=0;tempaddress<size;tempaddress++)
            buffer[tempaddress] = recchar();

        /* Then program the EEPROM */
        _WAIT_FOR_SPM();
    	for( tempaddress=0; tempaddress < size; tempaddress++)
    	{
	        EEARL = *address; // Setup EEPROM address
            EEARH = ((*address) >> 8);
            EEDR = buffer[tempaddress]; // Get byte.
            EECR |= (1<<EEMWE); // Write byte.
            EECR |= (1<<EEWE);
            while (EECR & (1<<EEWE)) // Wait for write operation to finish.
                ;

  			(*address)++; // Select next EEPROM byte
        }

        return '\r'; // Report programming OK
    }

    // Flash memory type.
	else if(mem=='F')
    { // NOTE: For flash programming, 'address' is given in words.
# ifdef LARGE_MEMORY
        (*address) <<= 2; // Convert address to bytes temporarily.
# else
        (*address) <<= 1; // Convert address to bytes temporarily.
# endif
        tempaddress = (*address);  // Store address in page.
	
        do
		{
            data = recchar();
            data |= (recchar() << 8);
            #pragma diag_suppress=Pe1053 // Suppress warning for conversion from long-type address to flash ptr.
            _FILL_TEMP_WORD(*address,data);
            #pragma diag_default=Pe1053 // Back to default.
# ifdef LARGE_MEMORY
            (*address)+=3; // Select next word in memory.
# else
            (*address)+=2; // Select next word in memory.
# endif
            size -= 2; // Reduce number of bytes to write by two.
        } while(size); // Loop until all bytes written.

	#pragma diag_suppress=Pe1053 // Suppress warning for conversion from long-type address to flash ptr.
	_PAGE_WRITE(tempaddress);
	#pragma diag_default=Pe1053 // Back to default.
	_WAIT_FOR_SPM();
	_ENABLE_RWW_SECTION();
# ifdef LERGE_MEMORY
        (*address) >>= 2; // Convert address back to Flash words again.
# else
        (*address) >>= 1; // Convert address back to Flash words again.
# endif
        return '\r'; // Report programming OK
    }

    // Invalid memory type?
    else
    {
        return '?';
    }
}


void BlockRead(unsigned int size, unsigned char mem, long *address)
{
    // EEPROM memory type.
    if (mem=='E') // Read EEPROM
    {
        do
        {
            EEARL = *address; // Setup EEPROM address
            EEARH = ((*address) >> 8);
            (*address)++; // Select next EEPROM byte
            EECR |= (1<<EERE); // Read EEPROM
            sendchar(EEDR); // Transmit EEPROM dat ato PC

            size--; // Decrease number of bytes to read
        } while (size); // Repeat until all block has been read
    }

    // Flash memory type.
	else if(mem=='F')
	{
# ifdef LARGE_MEMORY
        (*address) <<= 2; // Convert address to bytes temporarily.
# else
        (*address) <<= 1; // Convert address to bytes temporarily.
# endif
	
        do
        {
            #pragma diag_suppress=Pe1053 // Suppress warning for conversion from long-type address to flash ptr.
            sendchar( _LOAD_PROGRAM_MEMORY(*address) );
            sendchar( _LOAD_PROGRAM_MEMORY((*address)+1) );
# ifdef LARGE_MEMORY
            sendchar( _LOAD_PROGRAM_MEMORY((*address)+2) );
# endif
            #pragma diag_default=Pe1053 // Back to default.
# ifdef LARGE_MEMORY
            (*address) += 3; // Select next word in memory.
# else
            (*address) += 2; // Select next word in memory.
# endif
            size -= 2; // Subtract two bytes from number of bytes to read
        } while (size); // Repeat until all block has been read
# ifdef LARGE_MEMORY
        (*address) >>= 2; // Convert address back to Flash words again.
# else
        (*address) >>= 1; // Convert address back to Flash words again.
# endif
    }
}
#endif


/* end of file */
